# coding: utf-8
import sys
import logging
import argparse
from .client import OepClient
from .test import test as testscript
from . import __version__

logger = logging.getLogger()

# show version in help
class ArgumentParser(argparse.ArgumentParser):
    version_line = "oep-client version %s" % __version__

    def format_help(self):
        help = super().format_help()
        return self.version_line + "\n" + help


def setup_logging(loglevel="INFO"):
    if isinstance(loglevel, str):  # e.g. 'debug'/'DEBUG' -> logger.DEBUG
        loglevel = getattr(logging, loglevel.upper())
    formatter = logging.Formatter("[%(asctime)s %(levelname)7s] %(message)s")
    logger.setLevel(loglevel)

    # remove all existing handlers
    for h in logger.handlers:
        logger.removeHandler(h)

    h = logging.StreamHandler()
    h.setFormatter(formatter)
    h.setLevel(loglevel)
    logger.addHandler(h)


def _main(
    settings,
    metadata,
    validate,
    create,
    delete,
    upload_data,
    download_data,
    update_metadata,
    download_metadata,
    tablename,
    sheet,
    encoding,
    delimiter,
    token,
    test,
    test_rows,
    batch_size,
):
    # configure and validate settings
    if settings:
        settings = OepClient.load_json(settings)
    else:
        settings = {}
    if metadata:
        metadata = OepClient.load_json(metadata)
    else:
        metadata = {}
    if tablename:
        settings["tablename"] = tablename
    if token:
        settings["token"] = token

    cl = OepClient(**settings)
    if test:
        if any(
            (delete, validate, create, upload_data, download_data, download_metadata)
        ):
            raise Exception("Cannot use action in combination with test")
        testscript(token, test_rows=test_rows, batch_size=batch_size)
    elif delete:
        if any((test, validate, create, upload_data, download_data, download_metadata)):
            raise Exception("Cannot use action in combination with delete")
        cl.delete(metadata=metadata)
    elif download_data or download_metadata:  # download workflow
        if any((test, delete, validate, create, upload_data, update_metadata)):
            raise Exception(
                "Cannot use action in combination with download_data/download_metadata"
            )
        if download_data:
            df = cl.download_data(metadata=metadata)
            cl.save_dataframe(
                df, download_data, sheet=sheet, delimiter=delimiter, encoding=encoding
            )
        if download_metadata:
            meta = cl.download_metadata(metadata=metadata)
            cl.save_json(meta, download_metadata)
    elif create or upload_data or update_metadata or validate:  # upload workflow
        if any((test, delete, download_data, download_metadata)):
            raise Exception(
                "Cannot use action in combination with create/upload_data/update_metadata"
            )
        if validate or create or update_metadata:
            if not metadata:
                raise Exception("upload/validate requires metadata file")
        if validate:
            cl.validate(metadata)
        if create:
            cl.create(metadata)
        if upload_data:
            df = cl.load_dataframe(
                upload_data, sheet=sheet, delimiter=delimiter, encoding=encoding
            )
            cl.upload_data(df, metadata=metadata, batch_size=batch_size)
        if update_metadata:
            cl.update_metadata(metadata)
    else:
        logger.warning("No action selected")


def main():
    # arguments
    ap = ArgumentParser()
    ap.add_argument("--settings", "-s", help="json file with additional configuration")
    ap.add_argument("--metadata", "-m", help="json file with meta data")
    ap.add_argument("--token", "-t", help="user token")
    ap.add_argument(
        "--validate", action="store_true", help="validate meta data before upload"
    )
    ap.add_argument(
        "--create", "-c", action="store_true", help="create table at the beginning"
    )
    ap.add_argument("--delete", action="store_true", help="remove table")
    ap.add_argument(
        "--upload-data", "-u", help="input file with data, can be xlsx, csv, json"
    )
    ap.add_argument(
        "--download-data", "-d", help="file to save data to, can be xlsx, csv, json"
    )
    ap.add_argument(
        "--update-metadata", action="store_true", help="update metadata on platform"
    )
    ap.add_argument(
        "--download-metadata", help="download metadata from platform and save it (json)"
    )
    ap.add_argument("--tablename", "-n", help="table name")
    ap.add_argument("--sheet", help="sheet name, only for xlsx files")
    ap.add_argument("--encoding", "-e", help="text file encoding, only csv")
    ap.add_argument("--delimiter", help="field sparator, only csv")
    ap.add_argument(
        "--loglevel",
        "-l",
        default="INFO",
        type=str,
        help="ERROR, WARNING, INFO, or DEBUG",
    )
    ap.add_argument("--test", action="store_true", help="run test script")
    ap.add_argument("--test_rows", help="number of test rows to upload", type=int)
    ap.add_argument(
        "--batch_size",
        help="number of records to upload in one batch. 0 = all at once, None = default value",
        type=int,
    )
    kwargs = vars(ap.parse_args())
    setup_logging(kwargs.pop("loglevel"))

    rc = 0
    try:
        _main(**kwargs)
    except KeyboardInterrupt:
        rc = 130
    except Exception as e:
        if logger.level == logging.DEBUG:
            logger.error(e, exc_info=e)  # show error trace
        else:
            logger.error(e)
        rc = 1
    sys.exit(rc)


if __name__ == "__main__":
    main()



# coding: utf-8

import re
import json
import requests
import logging
import pandas as pd
from numpy import nan
import omi.dialects.oep.parser

logger = logging.getLogger()


def fix_name(name):
    name_new = re.sub("[^a-z0-9_]+", "_", name.lower())
    if name_new != name:
        logger.warning('Changed name "%s" to "%s"' % (name, name_new))
    return name_new


class OepClient:
    def __init__(self, **settings):
        self.settings = {
            "host": "openenergy-platform.org",
            "api_version": "v0",
            "protocol": "https",
            "schema_draft": "model_draft",
            "schema": "model_draft",
            "batch_size": 1000,
            "encoding": "utf-8",
            "delimiter": ",",
            "ssl_verify": True,
            "metadata_version": "1.4",
        }
        self.settings.update(settings)
        for k, v in self.settings.items():
            logger.debug("%s = %s" % (k, v))
        self.session = requests.session()
        self.session.verify = self.settings["ssl_verify"]
        if "token" in self.settings:
            self.session.headers = {"Authorization": "Token %(token)s" % self.settings}

    def request(self, method, url, jsondata=None):
        logger.info("%s %s" % (method, url))
        res = self.session.request(url=url, method=method, json=jsondata)
        try:
            res.raise_for_status()
        except Exception as e:
            try:
                err = res.json()
            except:
                err = e
            raise Exception(err)
        return res

    def validate(self, metadata):
        """
        """
        logger.info("VALIDATE")
        validator = getattr(
            omi.dialects.oep.parser,
            "JSONParser_" + self.settings["metadata_version"].replace(".", "_"),
        )()
        validator.parse(metadata)
        logger.info("   ok.")

    def delete(self, metadata=None):
        """
        """
        logger.info("DELETE")
        url = self.get_url(is_draft=True, metadata=metadata)
        self.request("DELETE", url)
        logger.info("   ok.")

    def create(self, metadata):
        """
        """
        logger.info("CREATE")
        url = self.get_url(is_draft=True, metadata=metadata)
        columns = self.get_column_defs_from_meta(metadata)
        jsondata = {
            "query": {
                "columns": columns,
                # TODO: additional constraints from metadata
                "constraints": [
                    {"constraint_type": "PRIMARY KEY", "constraint_parameter": "id"}
                ],
            }
        }
        self.request("PUT", url, jsondata=jsondata)
        logger.info("   ok.")

    def upload_data(self, dataframe, metadata=None, batch_size=None):
        """
        """
        logger.info("UPLOAD_DATA")
        url = self.get_url(is_draft=True, metadata=metadata) + "rows/new"
        data = self.convert_dataframe(dataframe)
        n_records = len(data)
        if batch_size is None:  # use default
            batch_size = self.settings["batch_size"]
        elif batch_size == 0:
            batch_size = n_records
        while data:
            batch, data = data[:batch_size], data[batch_size:]
            jsondata = {"query": batch}
            logger.info("   sending batch (n=%d) ..." % len(batch))
            self.request("POST", url, jsondata=jsondata)
        logger.info("   ok. (n=%d)" % n_records)

    def download_data(self, metadata=None):
        """
        """
        logger.info("DOWNLOAD_DATA")
        url = self.get_url(is_draft=False, metadata=metadata) + "rows/"
        res = self.request("GET", url)
        try:
            # TODO: file bugreport: empty table returns invalid json (content = b']')
            res = res.json()
        except Exception as e:
            logger.warning("Empty table")
            res = []
        df = pd.DataFrame(res)
        # df.set_index(['id'], inplace=True)  # make id column to index
        logger.info("   ok. (n=%d)" % len(res))
        return df

    def update_metadata(self, metadata):
        """
        """
        logger.info("UPDATE_METADATA")
        url = self.get_url(is_draft=True, metadata=metadata) + "meta/"
        self.request("POST", url, jsondata=metadata)
        logger.info("   ok.")

    def download_metadata(self, metadata=None):
        """
        """
        logger.info("DOWNLOAD_METADATA")
        url = self.get_url(is_draft=False, metadata=metadata) + "meta/"
        res = self.request("GET", url)
        res = res.json()
        logger.info("   ok.")
        return res

    @classmethod
    def save_dataframe(
        cls, df, filepath, sheet=None, delimiter=",", encoding="utf-8", index=False
    ):
        """
        """
        if filepath.endswith(".xlsx"):
            if not sheet:
                raise Exception("Must specify sheet when using xlsx")
            df.to_excel(filepath, sheet_name=sheet, index=index)
        elif filepath.endswith(".csv"):
            df.to_csv(
                filepath, encoding=encoding, sep=delimiter, na_rep="", index=index
            )
        elif filepath.endswith(".json"):
            data = cls.convert_dataframe(df)
            cls.save_json(data, filepath)
        else:
            raise Exception("Unsupported data type: %s" % filepath)
        return df

    @classmethod
    def convert_dataframe(cls, df):
        columns = [fix_name(n) for n in df.columns]
        # replace nan
        df = df.replace({nan: None})
        data = df.values.tolist()
        data = [dict(zip(columns, row)) for row in data]
        return data

    @classmethod
    def load_dataframe(cls, filepath, sheet=None, delimiter=",", encoding="utf-8"):
        """
        """
        if filepath.endswith(".xlsx"):
            if not sheet:
                raise Exception("Must specify sheet when using xlsx")
            df = pd.read_excel(filepath, sheet_name=sheet)
        elif filepath.endswith(".csv"):
            df = pd.read_csv(
                filepath,
                encoding=encoding,
                sep=delimiter,
                na_values=[""],
                keep_default_na=False,
            )
        elif filepath.endswith(".json"):
            df = pd.read_json(filepath, orient="records")
        else:
            raise Exception("Unsupported data type: %s" % filepath)
        return df

    @staticmethod
    def load_json(filepath):
        logger.info("reading %s" % filepath)
        with open(filepath, "rb") as f:
            return json.load(f)

    @staticmethod
    def save_json(data, filepath, encoding="utf-8"):
        logger.info("saving %s" % filepath)
        with open(filepath, "w", encoding=encoding) as f:
            return json.dump(data, f, sort_keys=True, indent=2)

    @staticmethod
    def get_column_defs_from_meta(metadata):
        """Return column definitions as list of {
            name: STR,
            type: STR,
            [character_maximum_length: INT],
            [is_nullable:YES|NO],
            [unit],
            [description]
        }"""
        res = []
        for c in metadata["resources"][0]["schema"]["fields"]:
            f = {
                "name": fix_name(c["name"]),
                "data_type": fix_name(c["type"]),
                "is_nullable": c.get("is_nullable", "YES"),
                "description": c.get("description", ""),
                "unit": c.get("unit", ""),
            }
            if "character_maximum_length" in c:
                f["character_maximum_length"] = int(c["character_maximum_length"])
            # fix type names
            if f["data_type"] == "double precision":
                f["data_type"] = "float"
            elif f["data_type"] == "serial":
                f["data_type"] = "integer"
            elif f["data_type"] == "string":
                f["data_type"] = "varchar"
            res.append(f)
        # add id, if not exist
        if not any(c["name"].lower() == "id" for c in res):
            res = [{"name": "id", "data_type": "bigserial", "is_nullable": "NO"}] + res
        return res

    def get_tablename_from_meta(self, metadata):
        try:
            name = metadata["resources"][0]["name"]
        except:
            raise Exception("table name not found in metadata (name in resource[0])")
        return name

    def get_url(self, is_draft=False, metadata=None):
        tablename = self.settings.get("tablename") or self.get_tablename_from_meta(
            metadata
        )
        if "." in tablename:
            schema, tablename = tablename.split(".")
        else:
            schema = self.settings["schema"]
        if is_draft:
            schema = self.settings["schema_draft"]
        url = "%(protocol)s://%(host)s/api/%(api_version)s/schema/" % self.settings
        url += "%s/tables/%s/" % (schema, tablename)
        return url



# coding: utf-8
import datetime
from .client import OepClient
import pandas as pd
import random
from argparse import ArgumentParser


example_metadata = {
    "id": "test",
    "description": "Test Table",
    "keywords": ["test"],
    "resources": [
        {
            "name": "test",
            "schema": {
                "fields": [
                    {
                        "name": "field1",
                        "type": "string",
                        "description": "column description",
                    },
                    {"name": "field2", "type": "integer", "unit": "none"},
                ],
                "foreignKeys": [],
            },
        }
    ],
}

exacmple_record = {"field1": "test", "field2": 999}


def test(token, test_rows=10, batch_size=None):
    example_data = pd.DataFrame([exacmple_record] * test_rows)
    timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    tablename = "test_%s_%d" % (timestamp, random.randint(100000, 999999))
    cl = OepClient(token=token, tablename=tablename)
    cl.create(metadata=example_metadata)
    try:
        cl.upload_data(dataframe=example_data, batch_size=batch_size)
        cl.update_metadata(metadata=example_metadata)
        data = cl.download_data()
        metadata = cl.download_metadata()
        print(len(data))
        print(data[0])
        print(metadata)
    finally:
        cl.delete()



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="oep-client",
    version="0.4.0",
    description="client side tool for openenergy platform",
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=["oep_client"],
    author="Christian Winger",
    author_email="c.winger@oeko.de",
    url="https://github.com/wingechr/oep-client",
    install_requires=["requests", "pandas", "omi"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["oep-client = oep_client.script:main"]},
    package_data={
        # 'package.module: [file_patterns]'
    },
)
